/////////////////////////////////////////////////////////////////////
// Validation Keycode
//
// This program calculates the Validation_Keycode for Templates for
// the IEEE 1451.4 Smart Transducer Standard.  The algorithm, as
// described in Section 7.2.5 is:
//
//   1. Initialize the 32-bit integer checksum value to 0.
//   2. For each line in the file up to but excluding the last 
//      line containing the Validation_Keycode command, do step 3.
//   3. For each ASCII character, including the carriage returns and 
//      line feeds, in the line do Steps 4-5.
//   4. Convert the ASCII character to an unsigned 8-bit integer.
//   5. Add the 8-bit value to the 32-bit checksum.
//
// Note that it is not clear as to whether the checksum integer
// shoud be signed or unsigned.  Since it is not stated in the
// the standard, it is assumed to be unsigned here.
//
// This program will return the proper keycode and inform you if
// it does not match the keycode in the template file.
//
// This is a completely different program from the one found in the
// Annex D of the standard. Here are some notes about that code:
//
//   1. There is a potential parsing error.  It will return an 
//      error if there is a single / inside an Abstract statement.  
//      Thus, for example, it will not properly parse the standard 
//      TEMPLATE 0,8,34,
//      "AC Linear/Rotary Variable Differential Transformer (LVDT/RVDT)"
//      found in Annex A.
//   2. The code uses type 'int' for the checksum.  The size of this is 
//      system dependent in most implementations of C and C++.  Thus it 
//      may not be the required 32 bits. In particular, if this is
//      implemented on a 16 bit machine, the checksum will be different
//      for a file containing all standard templates as found in Annex A,
//      than if it is implemented on a 32 bit machine because the checksum
//      exceeds 2^16.
//   4. The checksum will not be calculated correctly or an error is 
//      returned if "validation_keycode" is in an Abstract statement.
//   5. It will recognize a Validation_Keycode statement if it is on 
//      the same line as the ENDTEMPLATE statement, which is technically
//      not legal.  Further, in this situation, it does not include all
//      the characters upto the Validation_Keycode statement.  Thus,
//      the calculated checksum is not truly unique in this situation.
//
// Limited testing indicates that, for the cases when the code in the 
// standard operates correctly, it provides the same checksum as this code.
//
// There are some general issues related to the checksum in terms of
// intersystem compatibility.
//
//   1. There is no universally used extended version of the
//      ASCII character set.  Thus, any characters beyond the
//      standard 7-bit character set may not translate properly
//      across systems.  This is especially true internationally.
//      This may or may not affect the checksum.  Note that some
//      of the standard templates use a superscripted 2 as an
//      exponent.  This character is part of an extended ASCII set.
//   2. Some systems use carriage returns and line feeds at the
//      end of lines, whereas some only use line feeds.  Adding or
//      removing carriage returns will change the checksum.  
//   3. A difference in big endian or little endian may affect
//      the checksum.
//
// Coding Environment
// ------------------
//
// This code was originally written on Windows 2000 using Microsoft
// Visual C++ Version 6. A reasonable attempt was made to make the
// code generic so that it should be platform independent, but
// there are no guarentees to this effect.  
//
// WARRANTY DISCLAIMER. 
// --------------------
// There are no guarantees of any kind associated with use of this 
// code and the user uses it at their own risk.
//
// Running the program.
// --------------------
//
// The design is really meant to be run from a command prompt with
// the input template file as the one and only argument.  For example:
//
//   >validation_keycode ext_funct.tdl
//
// Alternatively, if it is run without a command line argument, or if
// it is run by double clicking on the executable in Windows, then
// you will be prompted for the input template file.  Hitting return
// at the prompt will cause the program to read from the file default.tdl.
//
// A file called messages.dat is created containing messages.
//
/////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define MAX_STRING 2000

FILE *messages;			// output file with all messages

/////////////////////////////////////////////////////////////////////
// convert_to_upper
//
// Converts a line to all upper case.
// 
// Input: line
// Output: none
// Return Value: converted line.
//
/////////////////////////////////////////////////////////////////////
void convert_to_upper (char* line, char* upper_line)
{

  for (int index = 0; line[index] != '\0'; index++) {
    upper_line[index] = toupper(line[index]);
  }

  upper_line[index] = '\0';

} // end convert_to_upper

/////////////////////////////////////////////////////////////////////
// validate_keycode_statement
//
// Checks to see if the current line contains a Validation_Keycode
// statement.  Since the Validation_Keycode statement must be the
// last line of the template file (see item 2 of the algorithm),
// it is only necessary to check to see that the first non whitespace
// of the line is 'Validation_Keycode'.  If it is, then pull out the
// keycode.
//
// Input: line - The input line from the template file.
//
// Output: keycode_read - the value of the keycode if a
//           Validation_Keycode statement is found.
//         An error message is output if the value associated
//           with the keycode is not an integer.
//
// Return value:
//
//  True if a Validation_Keycode statement is found.
//  False if it is not found.
//
/////////////////////////////////////////////////////////////////////
bool validate_keycode_statement (char* line, unsigned long* keycode_read)
{

  char keyword[MAX_STRING];
  char upper_case_keyword[MAX_STRING];
  int valid_keycode = 0;
  
  if ((valid_keycode = sscanf (line, "%s %d", keyword, keycode_read)) > 0) {

    convert_to_upper(keyword, upper_case_keyword); // TDL is case independent.

    if (strcmp(upper_case_keyword, "VALIDATION_KEYCODE") == 0) {

      if (valid_keycode != 2) {
        printf ("\nERROR: Validation Keycode not an integer.\n");
      }
      return (true);

    }else{

      *keycode_read = 0;
      return (false);

    } // end if-else

  }else{

    *keycode_read = 0;
    return (false);

  } // end if-else

} //end validate_keycode_statement

/////////////////////////////////////////////////////////////////////
// calculate_keycode
//
// Inputs: temmplate_file - the tdl file to be checked
//
// Outputs: calculated_keycode
//          read_keycode
//
// The algorithm implementation is as follows:
//
//  1. Read in a line from the file.
//  2. Check if that line has a Validation_Keycode statement.
//  3. Depending on step 2, loop through the characters in the line
//     adding to the keycode, or terminate.
//
/////////////////////////////////////////////////////////////////////
void calculate_keycode (FILE *template_file, unsigned long *calculated_keycode, 
                        unsigned long *keycode_read)
{
  char line[MAX_STRING];
  bool keycode_found = false;
  int line_count = 0;

  ///////////////////////////////////////////
  // Loop through each line.
  ///////////////////////////////////////////

  while (   fgets (line, MAX_STRING, template_file) != NULL 
         && (keycode_found = validate_keycode_statement(line, keycode_read)) == false ) {

    printf ("%d\n", ++line_count);
    //printf ("%d: %s", line_count, line);
    //fprintf (messages, "%d: %s", line_count, line);
    
    // Loop through each character in the line.

    int index=0; 
    while (line[index] != '\0') {

      *calculated_keycode += (unsigned char) line[index];
      index++;

    } // end while

  } // end while

  ///////////////////////////////////////////
  // Check if keycode valid.
  ///////////////////////////////////////////

  if (keycode_found) {

    if (*calculated_keycode == *keycode_read) {
      printf ("\nKeycode CORRECT.\n\n");
    }else{
      printf ("\nWARNING: Keycode NOT CORRECT.\n\n");
    } // end if-else

  }else{
    
    printf ("\nERROR: No Validation_Keycode found.\n\n");

  } // end if-else

} // end calculate_keycode

/////////////////////////////////////////////////////////////////////
// Main function
/////////////////////////////////////////////////////////////////////
void main (int argc, char **argv)

{
  char filename[100]; 
  FILE *template_file;
  unsigned long calculated_keycode = 0; 
  unsigned long keycode_read = 0;

  /////////////////////////////////////////
  // Open the messages file
  /////////////////////////////////////////

  strcpy (filename, "message.dat");
  if ( (messages=fopen(filename, "w")) == NULL) {
    printf ("ERROR: TDL Validation Keycode: Can not open file %s\n", filename);
    fprintf (messages, "ERROR: TDL Validation Keycode: Can not open file %s\n", filename);
    printf ("Hit any key to exit.");
    getchar();
    exit(1);
  };

  printf ("TDL Validation Keycode program.\n");
  fprintf (messages, "TDL Validation Keycode program.\n");

  /////////////////////////////////////////
  // Open the input file
  /////////////////////////////////////////

  if (argc == 1) {
    printf ("Template file name [default.tdl]?");
    gets (filename);
    if (strcmp (filename, "") == 0) {
      strcpy (filename, "default.tdl");
    }
  }else{
    strcpy (filename, argv[1]);
  };

  if ( (template_file=fopen(filename, "r")) == NULL) {
    printf ("ERROR: TDL Validation Keycode: Can not open file %s\n", filename);
    fprintf (messages, "ERROR: TDL Validation Keycode: Can not open file %s\n", filename);
    printf ("Hit any key to exit.");
    getchar();
    exit(1);
  };

  printf ("Reading from file '%s'.\n", filename);
  fprintf (messages, "Reading from file '%s'.\n", filename);

  //////////////////////////////////////////////
  // parse the file to calculate the keycode
  //////////////////////////////////////////////

  calculate_keycode (template_file, &calculated_keycode, &keycode_read);

  printf ("Calculated Keycode = %d\n", calculated_keycode);
  printf ("Keycode read = %ld\n", keycode_read);
  fprintf (messages, "Calculated Keycode = %d\n", calculated_keycode);
  fprintf (messages, "Keycode read = %ld\n", keycode_read);

  //////////////////////////////////////////////
  // Clean up
  //////////////////////////////////////////////

  fprintf (messages, "\nDone\n");
  fclose (messages);
  printf ("Done\n");

  printf ("Hit any key to exit.");
  getchar();

} // end main